@@ -11,6 +11,8 @@ function(i, j, v, nrow = max(i), ncol = max(j), dimnames = NULL)
                 nrow = as.integer(nrow), ncol = as.integer(ncol),
                 dimnames = dimnames)
     class(stm) <- "simple_triplet_matrix"
+    if(!.Call(R__valid_stm, stm))
+	stop("failed to create a valid 'simple_sparse_matrix' object")
     stm
 }
 
@@ -23,6 +25,7 @@ as.simple_triplet_matrix.simple_triplet_matrix <- identity
 as.simple_triplet_matrix.matrix <-
 function(x)
 {
+    x <- unclass(x)
     if(!prod(dim(x)))
         return(simple_triplet_matrix(integer(), integer(), c(x),
                                      nrow = nrow(x), ncol = ncol(x),
@@ -36,7 +39,9 @@ function(x)
 
 as.simple_triplet_matrix.default <-
 function(x)
-    as.simple_triplet_matrix(as.matrix(x))
+    as.simple_triplet_matrix(unclass(as.matrix(x)))
+
+## Sparse matrix classes in package 'Matrix'.
 
 as.simple_triplet_matrix.dgTMatrix <-
 function(x)
@@ -49,13 +54,57 @@ as.simple_triplet_matrix.dgCMatrix <-
 function(x)
 {
     nc <- x@Dim[2L]
-    simple_triplet_matrix(x@i + 1L, rep.int(seq_len(nc), diff(x@p)), x@x,
+    simple_triplet_matrix(x@i + 1L, rep.int(seq_len(nc), diff(x@p)),
+                          x@x,
                           x@Dim[1L], nc, x@Dimnames)
 }
 
+as.simple_triplet_matrix.dgRMatrix <-
+function(x)
+{
+    nr <- x@Dim[1L]
+    simple_triplet_matrix(rep.int(seq_len(nr), diff(x@p)), x@j + 1L,
+                          x@x,
+                          nr, x@Dim[2L], x@Dimnames)
+}
+
 ## See Work/Matrix.R for S4 methods for coercing simple triplet matrices
 ## to Matrix objects.
 
+## Sparse matrix classes in package 'SparseM'.
+
+as.simple_triplet_matrix.matrix.coo <-
+function(x)
+    simple_triplet_matrix(x@ia, x@ja, x@ra,
+                          x@dimension[1L], x@dimension[2L])
+
+as.simple_triplet_matrix.matrix.csc <-
+function(x)
+{
+    nc <- x@dimension[2L]
+    simple_triplet_matrix(x@ja, rep.int(seq_len(nc), diff(x@ia)), x@ra,
+                          x@dimension[1L], nc)
+}
+
+as.simple_triplet_matrix.matrix.csr <-
+function(x)
+{
+    nr <- x@dimension[1L]
+    simple_triplet_matrix(rep.int(seq_len(nr), diff(x@ia)), x@ja, x@ra,
+                          nr, x@dimension[2L])
+}
+
+## Sparse matrix class in package 'spam'.
+
+as.simple_triplet_matrix.spam <-
+function(x)
+{
+    nr <- x@dimension[1L]
+    simple_triplet_matrix(rep.int(seq_len(nr), diff(x@rowpointers)),
+                          x@colindices, x@entries,
+                          nr, x@dimension[2L])
+}
+
 as.matrix.simple_triplet_matrix <-
 function(x, ...)
 {
@@ -67,14 +116,66 @@ function(x, ...)
     y
 }
 
+as.array.simple_triplet_matrix <-
+function(x, ...)
+    as.array(as.matrix.simple_triplet_matrix(x, ...))
+
+as.simple_triplet_matrix.simple_sparse_array <-
+function(x) {
+    dx <- x$dim
+    if(length(dx) == 1L) {
+        simple_triplet_matrix(
+	    i = x$i[, 1L],
+            j = rep.int(1L, nrow(x$i)),
+            v = x$v,
+            nrow = dx,
+            ncol = 1L,
+            dimnames = 
+		if (!is.null(x$dimnames))
+		    c(x$dimnames, list(NULL))
+		else
+		    NULL
+	)
+    } 
+    else 
+    if(length(dx) == 2L) {
+	simple_triplet_matrix(
+	    i = x$i[,1L],
+	    j = x$i[,2L],
+	    v = x$v,
+	    nrow = x$dim[1L],
+	    ncol = x$dim[2L],
+	    dimnames = x$dimnames
+	)
+    }
+    else
+	stop("Unsupported number of dimensions")
+}
+
+
 is.simple_triplet_matrix <-
 function(x)
     inherits(x, "simple_triplet_matrix")
 
+is.numeric.simple_sparse_array <-
 is.numeric.simple_triplet_matrix <-
 function(x)
     is.numeric(x$v)
 
+Math.simple_triplet_matrix <-
+function(x, ...)
+{
+    ## Functions in the Math group mapping 0 to 0:
+    funs <- c("abs", "sign", "sqrt",
+              "floor", "ceiling", "trunc", "round", "signif")
+    if(is.na(match(as.character(.Generic), funs)))
+        stop(gettextf("Generic '%s' not defined for \"%s\" objects.",
+                      .Generic, .Class),
+             domain = NA)
+
+    x$v <- get(.Generic)(x$v, ...)
+    x
+}
 
 Ops.simple_triplet_matrix <-
 function(e1, e2)
@@ -122,23 +223,26 @@ function(e1, e2)
             return(e1)
         }
         stop(gettextf("Unary '%s' not defined for \"%s\" objects.",
-                      .Generic, .Class))
+                      .Generic, .Class),
+             domain = NA)
     }
 
     if(!(op %in% c("+", "-", "*", "/", "^",
                    "==", "!=", "<", "<=", ">", ">=")))
         stop(gettextf("Generic '%s' not defined for \"%s\" objects.",
-                      .Generic, .Class))
+                      .Generic, .Class),
+             domain = NA)
 
     ## Require numeric operands for the arithmetic operators.
     if(!is.numeric(e1) || !is.numeric(e2))
         stop("Not implemented.")
 
     if(op %in% c("==", "!=", "<", "<=", ">", ">=")) {
-        if(length(v <- e2) == 1L) {
-            if(is.na(v))
+        if(length(e2) == 1L) {
+            if(is.na(e2))
                 stop("NA/NaN handling not implemented.")
-            ind <- if(do.call(.Generic, list(0, v))) {
+	    names(e2) <- NULL
+            ind <- if(do.call(.Generic, list(0, e2))) {
                 ## This inverts the sparse storage advantage, and hence
                 ## will typically be inefficient.  Need to find the row
                 ## and column positions of the zero entries.
@@ -147,7 +251,7 @@ function(e1, e2)
                 which(m, arr.ind = TRUE)
             } else 
 		integer()
-            e1$v <- do.call(.Generic, list(e1$v, v))
+            e1$v <- do.call(.Generic, list(e1$v, e2))
 	    e1 <- .reduce(e1)
             if(n <- NROW(ind)) {
                 e1$i <- c(e1$i, ind[, 1L])
@@ -164,6 +268,7 @@ function(e1, e2)
         if(is.object(e2) || (length(e2) != 1L) ||
            !is.finite(e2) || (e2 <= 0))
             stop("Not implemented.")
+	names(e2) <- NULL
         e1$v <- e1$v ^ e2
         return(e1)
     }
@@ -196,24 +301,35 @@ function(e1, e2)
             if(length(e2) == 1L) {
 		if(!is.finite(e2))
 		    return(as.simple_triplet_matrix(as.matrix(e1) * e2))
+		names(e2) <- NULL
 		e1$v <- e1$v * e2
 		if(!e2)
 		    e1 <- .reduce(e1)
                 return(e1)
             }
             if(length(e2) == e1$nrow) {
+		names(e2) <- NULL
 		pos <- which(!is.finite(e2))
 		if(length(pos)) {
 		    ## replace with dense rows
 		    ind <- match(e1$i, pos, nomatch = 0L) == 0L
-		    e1$v <- c(e1$v[ind], as.matrix(e1[pos,]))
-		    e1$i <- c(e1$i[ind], rep.int(pos, e1$ncol))
-		    e1$j <- c(e1$j[ind], rep(seq_len(e1$ncol), 
-					     each = length(pos)))
+		    e1$v <- c(e1$v[ind],
+                              as.matrix(e1[pos, ]))
+		    e1$i <- c(e1$i[ind],
+                              rep.int(pos, e1$ncol))
+		    e1$j <- c(e1$j[ind],
+                              rep(seq_len(e1$ncol), each = length(pos)))
 		}
                 e1$v <- e1$v * e2[e1$i]
-		if (any(!e2))
+		if(any(!e2))
 		    e1 <- .reduce(e1)
+                ## Could add something like
+                ##    if(is.null(e1$dimnames) &&
+                ##       !is.null(nms <- names(e2))) {
+                ##        e1$dimnames <- list(nms, NULL)
+                ##    }
+                ## but then multiplying a matrix and a vector does not
+                ## seem to do this either ...
                 return(e1)
             }
 	    if(is.matrix(e2)) {
@@ -233,6 +349,7 @@ function(e1, e2)
 		e1$v <- e1$v * e2[cbind(e1$i, e1$j)]
 		if (any(!e2))
 		    e1 <- .reduce(e1)
+                e1$dimnames <- .make_dimnames(e1, e2)
 		return(e1)
 	    }
 	    stop("Not implemented.")
@@ -260,7 +377,7 @@ function(e1, e2)
 	    e2$i <- c(e2$i[ind], e2$i[-ind], e1$i[-pos])
 	    e2$j <- c(e2$j[ind], e2$j[-ind], e1$j[-pos])
 	    e2$v <- c(e2$v[ind] * e1$v[pos],
-		vector(typeof(e2$v), 1L) * c(e2$v[-ind], e1$v[-pos]))
+                      vector(typeof(e2$v), 1L) * c(e2$v[-ind], e1$v[-pos]))
 	    e2$dimnames <- .make_dimnames(e1, e2)
 	    return(.reduce(e2))
 	} else
@@ -328,32 +445,67 @@ dim.simple_triplet_matrix <-
 function(x)
     c(x$nrow, x$ncol)
 
-## <TODO>
-## Add a dim setter.
-## </TODO>
+`dim<-.simple_triplet_matrix` <-
+function(x, value)
+{
+    value <- as.integer(value)
+    if((length(value) != 2L) || any(is.na(value)))
+        stop("invalid dim replacement value")
+    nr <- x$nrow
+    nc <- x$ncol
+    if(prod(value) != nr * nc)
+        stop("invalid dim replacement value")
+
+    pos <- nr * (x$j - 1L) + x$i - 1L
+
+    nr <- value[1L]
+    nc <- value[2L]
+
+    x$i <- pos %%  nr + 1L
+    x$j <- pos %/% nr + 1L
+    x$nrow <- nr
+    x$ncol <- nc
+    x$dimnames <- NULL
+
+    x
+}
 
 dimnames.simple_triplet_matrix <-
 function(x)
     x$dimnames
 
-`dimnames<-.simple_triplet_matrix` <-
+`dimnames<-.simple_sparse_array` <-
+`dimnames<-.simple_triplet_matrix` <- 
 function(x, value)
 {
     if(!is.null(value)) {
-        ## Should be a list of length 2.
-        if(!is.list(value) || (length(value) != 2L))
+        ## NOTE that if length(value) < length(dim(x)) we
+	##      have to assume that the dimensions with index
+	##      seq_len(length(value)) are to be set. For 
+	##	example, we are called with a list of length
+	##      one if we call dimnames(x)[[1L]] <- value and
+	##      dimnames(x) == NULL (because of [[<-)
+	##
+        if(!is.list(value) || length(value) > length(dim(x)))
             stop("Invalid dimnames.")
-        ind <- sapply(value, length) == 0L
-        if(all(ind))
+        if(!length(value))
             value <- NULL
         else {
-            dnx <- vector("list", 2L)
-            dnx[!ind] <- lapply(value[!ind], as.character)
-            names(dnx) <- names(value)
+            dnx <- vector("list", length(dim(x)))
+	    len <- sapply(value, length)
+	    ind <- which(len > 0L)
+	    if (any(len[ind] != dim(x)[ind]))
+		stop("Invalid component length.")
+            dnx[ind] <- lapply(value[ind], as.character)
+	    if (!is.null(names(value))) {
+		ind <- seq_len(length(value))
+		names(dnx)[ind] <- names(value)
+	    }
         }
     }
+    ## See the constructor (above).
     if(is.null(value))
-        x["dimnames"] <- list(NULL)
+        x$dimnames <- NULL
     else
         x$dimnames <- dnx
     x
@@ -385,18 +537,38 @@ function(x, i, j, drop = FALSE)
         if(is.character(i))
             out <- vector(typeof(x$v))[rep.int(NA, length(i))]
         else if(!is.matrix(i)) {
-            if(is.logical(i))
-                i <- which(rep(i, length.out = nr))
+            if(is.logical(i)) {
+		if(nr * nc > 16777216L)
+		  stop("Logical vector subscripting disabled for this object.")
+                i <- which(rep(i, length.out = nr * nc))
+	    }
             else if(!is.numeric(i))
-                stop(gettextf("Invalid subscript type: %s.", typeof(i)))
+                stop(gettextf("Invalid subscript type: %s.",
+                              typeof(i)),
+                     domain = NA)
+	    else
+		## 52-bit safe
+		if(nr * nc > 4503599627370496)
+		  stop("Numeric vector subscripting disabled for this object.")
+	    ## Shortcut
+	    if(!length(i))
+		return(vector(mode = typeof(x$v), length = 0L))
+	    if(is.double(i))
+		i <- trunc(i)
             ## Let's hope we have a vector.
             ## What if we have both negatives and positives?
-            if(all(i >= 0)) {
+            if(all(i >= 0, na.rm = TRUE)) {
                 i <- i[i > 0]
                 out <- vector(mode = typeof(x$v), length = length(i))
-                pos <- match(i, (x$j - 1L) * nr + x$i, 0L)
-                out[pos > 0L] <- x$v[pos]
-            } else if(all(i <= 0)) {
+		if(length(out)) {
+		    is.na(i) <- i > nr * nc
+		    is.na(out) <- is.na(i)
+		    i <- match(i, (x$j - 1L) * nr + x$i, 0L)
+		    out[i > 0L] <- x$v[i]
+		}
+            } else if(!any(is.na(i)) && all(i <= 0)) {
+		if(nr * nc > 16777216L)
+		  stop("Negative vector subscripting disabled for this object.")
                 out <- vector(mode = typeof(x$v), nr * nc)
                 out[(x$j - 1L) * nr + x$i] <- x$v
                 out <- out[i]
@@ -404,23 +576,40 @@ function(x, i, j, drop = FALSE)
             else stop("Cannot mix positive and negative subscripts.")
         }
         else {
+	    ## Shortcut
+	    if(!nrow(i))
+		return(vector(mode = typeof(x$v), length = 0L))
+	    ## Ignore dimensions
+	    if(ncol(i) != 2L)
+		return(do.call("[.simple_triplet_matrix",
+			       list(x = x, as.vector(i))))
+
             ## Note that negative values are not allowed in a matrix
             ## subscript.
-            if((ncol(i) != 2L) || (any(i < 0)))
+	    if(is.double(i))
+		i <- trunc(i)
+            if(any(i < 0, na.rm = TRUE))
                 stop("Invalid subscript.")
             ## Rows containing zero indices can be dropped.
             ## Rows with NA indices should give NA (at least for
             ## non-recursive x).
-            i <- i[!apply(i == 0, 1L, any), , drop = FALSE]
+	    k <- .Call(R_all_row, i > 0, FALSE)
+            i <- i[k, ,drop = FALSE]
             out <- vector(mode = typeof(x$v), length = nrow(i))
-            ##  pi <- match(i[, 1L], x$i)
-            ##  pj <- match(i[, 2L], x$j)
-            ##  ind <- which(pi == pj)
-            ##  out[ind] <- x$v[pi[ind]]
-            pos <- match(paste(i[,1L], i[,2L], sep = "\r"),
-                         paste(x$i, x$j, sep = "\r"),
-                         nomatch = 0L)
-            out[pos > 0L] <- x$v[pos]
+	    if(length(out)) {
+		if (any(i > rep(c(nr, nc), each = nrow(i)), na.rm = TRUE))
+		    stop("subscript out of bounds")
+		k <- k[k]
+		is.na(out) <- is.na(k)
+		rm(k)
+		## See duplicated.matrix
+		## pos <- match(paste(i[, 1L], i[, 2L], sep = "\r"),
+		##	     paste(x$i, x$j, sep = "\r"),
+		##	     nomatch = 0L)
+		storage.mode(i) <- "integer"
+		i <- .Call(R_match_matrix, cbind(x$i, x$j), i, 0L)[[2L]]
+		out[i > 0L] <- x$v[i]
+	    }
         }
     }
     else {
@@ -430,6 +619,8 @@ function(x, i, j, drop = FALSE)
         ## columns must be unique.
         pos <- NULL
         if(!missing(i)) {
+	    if(any(is.na(i)))
+		stop("NA indices not allowed.")
             pi <- seq_len(nr)
             if(is.logical(i)) {
                 i <- rep(i, length.out = nr)
@@ -443,8 +634,12 @@ function(x, i, j, drop = FALSE)
                     if(any(duplicated(i)))
                         stop("Repeated indices currently not allowed.")
                 } else if(is.numeric(i)) {
+		    if(is.double(i))
+			i <- trunc(i)
                     if(all(i >= 0)) {
                         i <- i[i > 0]
+			if(any(i > nr))
+			    stop("subscript out of bounds")
                         if(any(duplicated(i)))
                             stop("Repeated indices currently not allowed.")
                     } else if(all(i <= 0))
@@ -453,7 +648,8 @@ function(x, i, j, drop = FALSE)
                         stop("Cannot mix positive and negative subscripts.")
                 } else {
                     stop(gettextf("Invalid subscript type: %s.",
-                                  typeof(i)))
+                                  typeof(i)),
+                         domain = NA)
                 }
                 nr <- length(i)
                 pos <- match(x$i, i, 0L) > 0L
@@ -461,6 +657,8 @@ function(x, i, j, drop = FALSE)
             pi[i] <- seq_len(nr)
         }
         if(!missing(j)) {
+	    if(any(is.na(j)))
+		stop("NA indices not allowed.")
             pj <- seq_len(nc)
             if(is.logical(j)) {
                 j <- rep(j, length.out = nc)
@@ -477,8 +675,12 @@ function(x, i, j, drop = FALSE)
                     if(any(duplicated(j)))
                         stop("Repeated indices currently not allowed.")
                 } else if(is.numeric(j)) {
+		    if(is.double(j))
+			j <- trunc(j)
                     if(all(j >= 0)) {
                         j <- j[j > 0]
+			if(any(j > nc))
+			    stop("subscript out of bounds")
                         if(any(duplicated(j)))
                             stop("Repeated indices currently not allowed.")
                     } else if(all(j <= 0))
@@ -487,7 +689,8 @@ function(x, i, j, drop = FALSE)
                         stop("Cannot mix positive and negative subscripts.")
                 } else {
                     stop(gettextf("Invalid subscript type: %s.",
-                                  typeof(j)))
+                                  typeof(j)),
+                         domain = NA)
                 }
                 nc <- length(j)
                 pos <- if(is.null(pos))
@@ -498,8 +701,20 @@ function(x, i, j, drop = FALSE)
             pj[j] <- seq_len(nc)
         }
 
-        if(!is.null(dnx <- x$dimnames))
-            dnx[] <- list(dnx[[1L]][i], dnx[[2L]][j])
+        if(!is.null(dnx <- x$dimnames)) {
+	    if (!missing(i)) {
+		dnx[1L] <- list(dnx[[1L]][i])
+		if (!length(dnx[[1L]]))
+		    dnx[1L] <- list(NULL)
+	    }
+	    if (!missing(j)) {
+		dnx[2L] <- list(dnx[[2L]][j])
+		if (!length(dnx[[2L]]))
+		    dnx[2L] <- list(NULL)
+	    }
+	    if (!length(dnx[[1L]]) && !length(dnx[[2L]]))
+		dnx <- NULL
+	}
 
         i <- if(missing(i)) x$i[pos] else pi[x$i[pos]]
         j <- if(missing(j)) x$j[pos] else pj[x$j[pos]]
@@ -526,8 +741,19 @@ function(..., deparse.level = 1L)
     }, args)
     ## Handle dimnames in one final step.
     rnms <- lapply(args, rownames)
-    rnms <- if(!any(sapply(rnms, is.null)))
+    rnms <- if(!all(sapply(rnms, is.null))) {
+	rnms <- mapply(function(rnm, n)
+	    if(is.null(rnm))
+		rep.int("", n)
+	    else
+		rnm,
+	    rnms,
+	    lapply(args, nrow),
+	    SIMPLIFY = FALSE
+	)
+	
         do.call("c", rnms)
+    }
     else
         NULL
     cnms <- Find(Negate(is.null), lapply(args, colnames))
@@ -551,12 +777,22 @@ function(..., deparse.level = 1L)
                               nrow = nr, ncol = nc + ncol(y))
     }, args)
     ## Handle dimnames in one final step.
-    rnms <- Find(Negate(is.null), lapply(args, rownames))
-    cnms <- lapply(args, rownames)
-    cnms <- if(!any(sapply(cnms, is.null)))
+    cnms <- lapply(args, colnames)
+    cnms <- if(!all(sapply(cnms, is.null))) {
+	cnms <- mapply(function(cnm, n)
+	    if(is.null(cnm))
+		rep.int("", n)
+	    else
+		cnm,
+	    cnms,
+	    lapply(args, ncol),
+	    SIMPLIFY = FALSE
+	)
         do.call("c", cnms)
+    }
     else
         NULL
+    rnms <- Find(Negate(is.null), lapply(args, rownames))
     dimnames(out) <- list(rnms, cnms)
     out
 }
@@ -620,8 +856,8 @@ function(..., recursive = FALSE)
 print.simple_triplet_matrix <-
 function(x, ...)
 {
-    writeLines(sprintf("A %s simple triplet matrix.",
-                       paste(dim(x), collapse = "x")))
+    writeLines(gettextf("A %s simple triplet matrix.",
+                        paste(dim(x), collapse = "x")))
     invisible(x)
 }
 
@@ -631,6 +867,29 @@ function(x, ...)
     sum(x$v) / prod(dim(x))
 }
 
+aperm.simple_triplet_matrix <-
+function(a, perm = NULL, ...)
+{
+    s <- c(1L, 2L)
+    if(!is.null(perm)) {
+        perm <- if(is.character(perm))
+            match(perm, names(a$dimnames))
+        else if(is.numeric(perm))
+            match(perm, s)
+        else NULL
+        if(length(perm) != length(s) || any(is.na(perm)))
+            stop("Invalid permutation.")
+        if(all(perm == s))
+            return(a)
+    }
+    ## Transpose.
+    t.simple_triplet_matrix(a)
+}
+
+as.vector.simple_triplet_matrix <-
+function(x, mode = "any")
+    as.vector(as.matrix(x), mode)
+
 ## Utilities for creating special simple triplet matrices:
 
 simple_triplet_zero_matrix <-
